#include <stdbool.h>


void initialize(long const N, long const M, __attribute__((unused)) long const TS, double (* restrict A)[M], double value) {
    for (long i = 0; i < N; i++) {
        for (long j = 0; j < M; j++) {
            A[i][j] = value;
        }
    }
}


void matmul(const long N, const long M, const long P, const long TS, double (*restrict A)[M/TS][TS][TS], double (*restrict B)[P/TS][TS][TS], double (*restrict C)[P/TS][TS][TS]) {
    long const NB_I = N / TS;
    long const NB_J = P / TS;
    long const NB_K = M / TS;

    for (long k = 0; k < NB_K; k++) {
        for (long i = 0; i < NB_I; i++) {
            for (long j = 0; j < NB_J; j++) {
                double (* restrict A_block)[TS] = A[i][k];
                double (* restrict B_block)[TS] = B[k][j];
                double (* restrict C_block)[TS] = C[i][j];

                for (long ii = 0; ii < TS; ii++) {
                    for (long jj = 0; jj < TS; jj++) {
                        for (long kk = 0; kk < TS; kk++) {
                            C_block[ii][jj] += A_block[ii][kk] * B_block[kk][jj];
                        }
                    }
                }
            }
        }
    }
}


bool verify(long const N, long const M, __attribute__((unused)) long const TS, double value, double (* restrict C)[M]) {
    bool validates = true;

    for (long i = 0; i < N; i++) {
        for (long j = 0; j < M; j++) {
            if (C[i][j] != value)
            {
                validates = false;
                break;
            }
        }
    }

    return validates;
}
